// Copyright (c) 2022 Marceline Cramer
// SPDX-License-Identifier: GPL-3.0-or-later

use crate::parse::ast;
use std::collections::HashMap;

type SymbolMap<T> = HashMap<String, T>;

#[rustfmt::skip]
#[derive(strum::Display, strum::EnumString)]
pub enum PrimitiveType {
    #[strum(serialize="i64")]    I64,
    #[strum(serialize="u64")]    U64,
    #[strum(serialize="i32")]    I32,
    #[strum(serialize="u32")]    U32,
    #[strum(serialize="i16")]    I16,
    #[strum(serialize="u16")]    U16,
    #[strum(serialize="i8")]     I8,
    #[strum(serialize="u8")]     U8,
    #[strum(serialize="f64")]    F64,
    #[strum(serialize="f32")]    F32,
    #[strum(serialize="string")] String,
    #[strum(serialize="bool")]   Bool,
}

#[derive(Clone, Debug)]
pub struct SymbolTable {
    pub structures: SymbolMap<StructDef>,
    pub interfaces: SymbolMap<IfaceDef>,
    pub functions: SymbolMap<FnDef>,
}

impl SymbolTable {
    pub fn from_ast<'a>(ast: &ast::Ast<'a>) -> Self {
        let mut structures = SymbolMap::new();

        for def in ast.defs.iter() {
            if let ast::Definition::Struct { name, members } = def {
                structures.insert(
                    name.to_string(),
                    StructDef {
                        members: members
                            .iter()
                            .map(|member| MemberDef {
                                member_name: member.name.to_string(),
                                type_name: member.type_name.to_string(),
                            })
                            .collect(),
                        interfaces: Vec::new(),
                        functions: Vec::new(),
                    },
                );
            }
        }

        let mut interfaces = SymbolMap::new();

        for def in ast.defs.iter() {
            if let ast::Definition::Interface { name, functions } = def {
                interfaces.insert(
                    name.to_string(),
                    IfaceDef {
                        functions: functions.iter().map(|f| f.into()).collect(),
                    },
                );
            }
        }

        let mut functions = SymbolMap::new();

        for def in ast.defs.iter() {
            if let ast::Definition::Function {
                associated_struct,
                implementation,
            } = def
            {
                let def: FnDef = (&implementation.def).into();
                if let Some(associated_struct) = associated_struct {
                    structures
                        .get_mut(&associated_struct.to_string())
                        .unwrap()
                        .functions
                        .push(def);
                } else {
                    functions.insert(def.name.to_owned(), def);
                }
            }
        }

        Self {
            structures,
            interfaces,
            functions,
        }
    }
}

#[derive(Clone, Debug)]
pub struct StructDef {
    pub members: Vec<MemberDef>,
    pub functions: Vec<FnDef>,
    pub interfaces: Vec<String>,
}

#[derive(Clone, Debug)]
pub struct MemberDef {
    pub member_name: String,
    pub type_name: String,
}

#[derive(Clone, Debug)]
pub struct IfaceDef {
    pub functions: Vec<FnDef>,
}

#[derive(Clone, Debug)]
pub struct FnDef {
    pub mutable: bool,
    pub name: String,
    pub args: Vec<FnArg>,
    pub return_type: Option<String>,
}

impl<'a> From<&ast::FnDef<'a>> for FnDef {
    fn from(def: &ast::FnDef<'a>) -> Self {
        Self {
            mutable: def.mutable,
            name: def.name.to_string(),
            args: def
                .signature
                .args
                .iter()
                .map(|arg| FnArg {
                    arg_name: arg.name.map(|n| n.to_string()),
                    type_name: arg.type_name.to_string(),
                })
                .collect(),
            return_type: def.signature.return_type.map(|t| t.to_string()),
        }
    }
}

#[derive(Clone, Debug)]
pub struct FnArg {
    pub arg_name: Option<String>,
    pub type_name: String,
}

#[cfg(test)]
mod tests {
    use super::*;

    fn from_source(source: &str) -> SymbolTable {
        use crate::parse::{rd, BuildAst};
        let ast = rd::RecursiveDescent::build_source_ast(source);
        let symtab = SymbolTable::from_ast(&ast);
        println!("{:#?}", symtab);
        symtab
    }

    #[test]
    fn from_clock() {
        from_source(include_str!("test/clock.fae"));
    }
}
